--------------------------BEGIN-VHDL-LICENSE-----------------------------
-- fasu.vhdl - Floating Point Add/Sub Execution Unit for the F-CPU
-- Copyright (C) 2003-2004 -- SEMET Gaetan <gaetan@xeberon.net>

-- This program is free software; you can redistribute it and/or modify
-- it under the terms of the GNU General Public License as published by
-- the Free Software Foundation; either version 2 of the License, or
-- (at your option) any later version.
--
-- This program is distributed in the hope that it will be useful,
-- but WITHOUT ANY WARRANTY; without even the implied warranty of
-- MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
-- GNU General Public License for more details.
--
-- You should have received a copy of the GNU General Public License
-- along with this program; if not, write to the Free Software
-- Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
---------------------------END-VHDL-LICENSE------------------------------


library IEEE;
use IEEE.std_logic_1164.all;
use IEEE.std_logic_arith.ALL;
use IEEE.std_logic_unsigned.ALL;

use std.textio.all; -- DEBUG ONLY
                  
use work.FCPU_config.all;
use work.generic_adder.all;
use work.Bit_Manipulation.all;

entity EU_FASU is
	generic (
		WIDTH : natural := 64;
		FLSIZE : natural := 32;
		--if (FLSIZE = 32) then
			MSIZE : natural := 23;
			ESIZE : natural := 8
		--end if;
	);

	port(
	-- inputs : 
		-- operands
		Din_0    : in std_ulogic_vector(WIDTH-1 downto 0);
		Din_1    : in std_ulogic_vector(WIDTH-1 downto 0);
		-- subtract flag (should be derived from opcode)
		-- 0 for fadd, 1 for fsub
		Substract : in std_ulogic;
		-- flag bits (directly copied from instruction word)
		Flags    : in std_ulogic_vector(23 downto 18);
		-- SIMD mode bits (decoded)
		SIMD     : in std_ulogic_vector(LOGMAXSIZE-1 downto 0);
		-- clock/reset/enable inputs
		Clk      : in std_ulogic;
		Rst      : in std_ulogic;
		En       : in std_ulogic;

	-- outputs:
		-- Result (8/16/32)
		Dout_0   : out std_ulogic_vector(WIDTH-1 downto 0);
		-- FPU Exception Flags
		FExout   : out std_ulogic_vector(8 downto 0)
		-- FExout(0) : DBZ  : Divide By Zero Detected
		-- FExout(1) : IOD  : Illegal Operand Detected
		-- FExout(2) : EUD  : Exponent Underflow Detected
		-- FExout(3) : EOD  : Exponent Overflow Detected
		-- FExout(4) : SNAN : Signaling Not A Number output
		-- FExout(5) : QNAN : Quiet Not A Number output
		-- FExout(6) : INF  : Infinite output
		-- FExout(7) : INE  : Inexact output
		-- FExout(8) : ZERO : Zero output 
	);
end EU_FASU;


-- Outputs:
--                +---------------------------------------------------+
--  Addition      |                    B                              |
--   A + B        +----------+----------+----------+---------+--------+
--                |    F     |  -infty  |  +infty  |  QNaN   |  SNaN  |
--+----+----------+----------+----------+----------+---------+--------+
--|    | F        |    F     |  -infty  |  +infty  |  QNaN   |  SNaN  |
--|    | -infty   |  -infty  |  -infty  |   QNaN   |  QNaN   |  SNaN  |
--| A  | +infty   |  +infty  |   QNaN   |  +infty  |  QNaN   |  SNaN  |
--|    | QNaN     |   QNaN   |   QNaN   |   QNaN   |  QNaN   |  SNaN  |
--|    | SNaN     |   SNaN   |   SNaN   |   SNaN   |  SNaN   |  SNaN  |
--+----+----------+----------+----------+----------+---------+--------+
--
--                +---------------------------------------------------+
--  Substraction  |                    B                              |
--   A - B        +----------+----------+----------+---------+--------+
--                |    F     |  -infty  |  +infty  |  QNaN   |  SNaN  |
--+----+----------+----------+----------+----------+---------+--------+
--|    | F        |    F     |  +infty  |  -infty  |  QNaN   |  SNaN  |
--| A  | -infty   |  -infty  |   QNaN   |  -infty  |  QNaN   |  SNaN  |
--|    | +infty   |  +infty  |  +infty  |   QNaN   |  QNaN   |  SNaN  |
--|    | QNaN     |   QNaN   |   QNaN   |   QNaN   |  QNaN   |  SNaN  |
--|    | SNaN     |   SNaN   |   SNaN   |   SNaN   |  SNaN   |  SNaN  |
--+----+----------+----------+----------+----------+---------+--------+
--
-- F : any finite floating value, ie except {S,Q}NaN, +/-infinity

-- Note FPU doesn't generate Signaling Not A Number (SNaN) except if
-- a SNaN was given in one operand.
-- Rem: IEEE wants a SNAN to be generated when overflow occurs, in place of
-- a QNAN
-- QNAN isgenerated when an illegal operation occurs


-- Operating mode:
--   Single: Float    : 32 bits
--           Mantissa : 23 bits
--           Exponent :  8 bits
--           Sign     :  1 bit
--   Double: Float    : 64 bits
--           Mantissa : 52 bits
--           Exponent : 11 bits
--           Sign     :  1 bit

architecture Behav_1 of EU_FASU is

-- COMPONENTS

	component IAdd is
		generic (
			WIDTH : natural := 64;
			PIPELINED : integer := 0
		);
		port (
			-- operand inputs
			A : in std_ulogic_vector(WIDTH-1 downto 0);
			B : in std_ulogic_vector(WIDTH-1 downto 0);
			-- subtract mode enable
			Sub : in std_ulogic;
			-- saturate/floor mode enable
			Sat : in std_ulogic;
			-- signed saturation mode
			Sig : in std_ulogic;
			-- increment (add) or decrement (sub) result by one
			Inc : in std_ulogic;
			-- average mode enable
			Avg : in std_ulogic;
			-- SIMD mode switches
			U : in std_ulogic_vector(2 downto 0);
			-- clock/reset inputs
			Clk : in std_ulogic;
			Rst : in std_ulogic;
			En : in std_ulogic;
		--
			-- 8-bit tap outputs
			Y8l : out std_ulogic_vector(WIDTH-1 downto 0);
			Y8h : out std_ulogic_vector(WIDTH-1 downto 0);
			-- regular outputs
			Yl : out std_ulogic_vector(WIDTH-1 downto 0);
			Yh : out std_ulogic_vector(WIDTH-1 downto 0)
		);
	end component;


	-- SIGNALS
	-- note : i use Symphony Sonata EDA Evaluation version for main programming
	-- and this version does not allow to waveform variable so I put lot of
	-- variables in signals...
	
	-- general signals
	signal Fa   : std_ulogic_vector (FLSIZE-1 downto 0);
	signal Fb   : std_ulogic_vector (FLSIZE-1 downto 0);
	signal Fout : std_ulogic_vector (FLSIZE-1 downto 0);	

	signal fl_ieee : std_ulogic;
	signal S1M : std_ulogic_vector(3 downto 0);

  -- stage 1 inputs
	signal S1Mode : std_ulogic_vector(3 downto 0);
	signal S1Fa   : std_ulogic_vector (FLSIZE-1 downto 0);
	signal S1Fb   : std_ulogic_vector (FLSIZE-1 downto 0);
	-- stage 1 outputs
	signal S1Fout  : std_ulogic_vector (FLSIZE-1 downto 0);	
	signal S1FExout: std_ulogic_vector(8 downto 0);
	signal S1out   : std_ulogic;	-- stage 1 have a result?

	-- stage 2 inputs
	signal S2Fa   : std_ulogic_vector (FLSIZE-1 downto 0);
	signal S2Fb   : std_ulogic_vector (FLSIZE-1 downto 0);
	signal S2Mode : std_ulogic_vector(3 downto 0);
	-- stage 2 signals
	signal S2ExpAddEn  : std_ulogic := '0';
	signal S2ExpAddOut : F_VECTOR;
	signal S2Ea: std_ulogic_vector (ESIZE-1 downto 0);
	signal S2Eb: std_ulogic_vector (ESIZE-1 downto 0);
	signal S2EaF: F_VECTOR;
	signal S2EbF: F_VECTOR;
	-- stage 2 outputs
	signal S2Fout  : std_ulogic_vector (FLSIZE-1 downto 0);	
	signal S2FExout: std_ulogic_vector(8 downto 0);
	signal S2out   : std_ulogic;

	-- stage 3 inputs
	signal S3Fa   : std_ulogic_vector (FLSIZE-1 downto 0);
	signal S3Fb   : std_ulogic_vector (FLSIZE-1 downto 0);
	signal S3Mode : std_ulogic_vector(3 downto 0);
	-- stage 3 signals
	signal S3DE : std_ulogic_vector (ESIZE downto 0);
  signal S3Ma : std_ulogic_vector (MSIZE downto 0);
  signal S3Mb : std_ulogic_vector (MSIZE downto 0);
  -- stage 3 outputs
	signal S3Fout  : std_ulogic_vector (FLSIZE-1 downto 0);	
	signal S3FExout: std_ulogic_vector(8 downto 0);
	signal S3out   : std_ulogic;
 
  -- stage 4 inputs
	signal S4Mode : std_ulogic_vector(3 downto 0);
  signal S4E : std_ulogic_vector(ESIZE-1 downto 0);
  signal S4Ma : std_ulogic_vector(FLSIZE-1 downto 0);
  signal S4Mb : std_ulogic_vector(FLSIZE-1 downto 0);
	signal S4MaF: F_VECTOR;
	signal S4MbF: F_VECTOR;
  signal S4Sa : std_ulogic;
  signal S4Sb : std_ulogic;
  signal S4AddSub : std_ulogic;
  --stage 4 signals	
	signal S4ManAddEn  : std_ulogic := '0';
	signal S4ManAddOut : F_VECTOR;
	-- stage 3 outputs
	signal S4Fout  : std_ulogic_vector (FLSIZE-1 downto 0);	
	signal S4FExout: std_ulogic_vector(8 downto 0);
	signal S4out   : std_ulogic;



  -- stage 5 inputs
	signal S5Mode : std_ulogic_vector(3 downto 0);
  signal S5E : std_ulogic_vector(ESIZE-1 downto 0);
	-- stage 5 signals
  signal S5M : std_ulogic_vector (MSIZE+2 downto 0);
  -- note : why MSIZE + 2?????
  -- S5M(MSIZE+2)    : to have sign
  -- S5M(MSIZE+1)    : in case of Mantissa Adder overflow
  -- S5M(MSIZE)      : MSB
  -- S5M(MSIZE-1..0) : 23 bit mantissa (for single)
  -- stage 6 outputs
	signal S5Fout : std_ulogic_vector (FLSIZE-1 downto 0);	
	signal S5FExout: std_ulogic_vector(8 downto 0);

	
  -- CONSTANTS

  -- indices for M Mode vector
	constant Mode_Sub : natural := 3;
	
	-- indices for FExout exception output vector
	constant Ex_DBZ  : natural := 0;  --  Divide By Zero Detected
	constant Ex_IOD  : natural := 1;  --  Illegal Operand Detected
	constant Ex_EUD  : natural := 2;  --  Exponent Underflow Detected
	constant Ex_EOD  : natural := 3;  --  Exponent Overflow Detected
	constant Ex_SNAN : natural := 4;  --  Signaling Not A Number output
	constant Ex_QNAN : natural := 5;  --  Quiet Not A Number output
	constant Ex_INF  : natural := 6;  --  Infinite output
	constant Ex_INE  : natural := 7;  --  Inexact output
	constant Ex_ZERO : natural := 8;  --  Zero output 


  -- special Floating values
  constant const32_infty   : std_ulogic_vector := "01111111100000000000000000000000"; --  +infinity
  constant const32_m_infty : std_ulogic_vector := "11111111100000000000000000000000"; --  -infinity
  constant const32_SNaN    : std_ulogic_vector := "01111111100000000000000000000001"; --    SNaN
  constant const32_QNaN    : std_ulogic_vector := "01111111110000000000000000000001"; --    QNaN
	constant const32_0       : std_ulogic_vector := "00000000000000000000000000000000"; --       0
	constant const32_m_0     : std_ulogic_vector := "10000000000000000000000000000000"; --      -0
  
  
-- INTERNAL FUNCTIONS
  procedure writestr (v : std_ulogic_vector) is
		variable lout : line;
		variable s: string(1 to v'length);
		variable m: STD_ULOGIC_VECTOR(1 to v'length) := v;
		type char_indexed_by_MVL9 is array (STD_ULOGIC) of character;
		constant MVL9_to_char: char_indexed_by_MVL9 := "UX01ZWLH-";
	begin
	  for i in 1 to v'length loop
			s(i) := MVL9_to_char(m(i));
		end loop;
		write(lout, s);
		-- rem: don't know why but write(lout,v) doesn't work:
		-- Error: CSVHEX0004: fasu.vhdl: (line 337): Overloaded Subprogram 'write' does not match its parameter profile
		writeline(output, lout);
	end writestr;

  procedure writestr (s : string) is
		variable lout : line;
	begin
		write(lout, s);
		writeline(output, lout);
	end writestr;

  -- this function right shift F by N bits 
  -- != from rshift from bit_manipulation.vhdl where N is a natural
  -- here N is unknown (it's a bit vector)
  -- used only for exponent right shifting
  -- exponent size is 8 bit for sigle float, and 11 for double float
  -- so we code N with 3 bits
  function fasu_rshift(F : std_ulogic_vector; N: std_ulogic_vector(2 downto 0)) return std_ulogic_vector is
    constant L : natural := F'length;
    variable yy1 : std_ulogic_vector(L-1 downto 0);
    variable yy2 : std_ulogic_vector(L-1 downto 0);
    variable yy3 : std_ulogic_vector(L-1 downto 0);
  begin
    if (N(0) = '1') then
      yy1(L-1) := '0';
      yy1(L-1-1 downto 0) := F(L-1 downto 1);
    else
      yy1 := F;
    end if;

    if (N(1) = '1') then
      yy2(L-1) := '0';
      yy2(L-2) := '0';
      yy2(L-1-2 downto 0) := yy1(L-1 downto 2);
    else
      yy2 := yy1;
    end if;
    
    if (N(2) = '1') then
      yy3(L-1 downto L-1-4) := (others => '0');
      yy3(L-1-4 downto 0) := yy2(L-1 downto 4);
    else
      yy3 := yy2;
    end if;

  	return yy3;
  end function;
  
  
  -- integer incrementer
  function fl_incr(F: std_ulogic_vector) return std_ulogic_vector is
  begin
    return F xor lshift(cascade_and(F), 1, '1');
  end function;

  -- integer decrementer
  function fl_decr(F: std_ulogic_vector) return std_ulogic_vector is
  begin
    return not fl_incr(not F);
  end function;

  -- note : use eu_inc instead of these functions???????????????????
	-- integer negation
	function fl_neg(F: std_ulogic_vector) return std_ulogic_vector is
  begin
    return fl_incr(not F);
  end function;
  

	-- return true is F only have 0
	function fl_isnull(F: std_ulogic_vector) return boolean is
		variable res : boolean;
		variable m   : std_ulogic_vector(1 to F'length) := F;
	begin
		res := true;
		for i in 1 to F'length loop
			if m(i) = '1' then
				res := false;
			end if;
		end loop;
		return res;
	end;

begin


-- USED COMPONENTS INSTANTIATIONS:

-- Exponent adder : 
	ExpAdder : IAdd
	  generic map (WIDTH => 64, PIPELINED => 1)
		port map (
			A => S2EaF,
			B => S2EbF,
			Sig => '0',
			inc => '0',
			avg => '0',
			Sub => '1',
			Sat => '0',
			U => "001", -- 16 bit SIMD for a 8 bit exponent (+ sign)
			Clk => not Clk,
			Rst => Rst,
			En => S2ExpAddEn,
			Yl => S2ExpAddOut
		);
		-- using existing IAdder in 16 bit SIMD mode
		-- but since A and B are 64 bit, there is 3 16bit
		-- useless substractions
		-- TODO: FIX IT
	S2EaF(WIDTH-1 downto ESIZE) <= (others => '0');
	S2EbF(WIDTH-1 downto ESIZE) <= (others => '0');

  -- result of stage 2 will be directly injected in stage 3
  S3DE <= S2ExpAddOut(ESIZE downto 0);
  
-- Mantissa adder
	ManAdder : IAdd
	  generic map (WIDTH => 64, PIPELINED => 1)
		port map (
			A => S4MaF,
			B => S4MbF,
			Sub => S4AddSub,
			Sat => '0',
			Sig => '0',
			inc => '0',
			avg => '0',
			U => "011", -- 32 bits SIMD for a 23 bit mantissa
			Clk => not Clk,
			Rst => Rst,
			En => S4ManAddEn,
			Yl => S4ManAddOut
		);
	-- note : here also an useless 32 add/sub operation (bits 63..32)
	S4MaF(WIDTH-1 downto FLSIZE) <= (others => '0');
	S4MbF(WIDTH-1 downto FLSIZE) <= (others => '0');

	S5M <= S4ManAddOut(MSIZE+2 downto 0);

-- FLOATING ADD/SUB UNIT STAGES : 


	Fa <= Din_0(FLSIZE-1 downto 0);
	Fb <= Din_1(FLSIZE-1 downto 0);
	Dout_0(WIDTH-1  downto FLSIZE) <= (others =>'0');
	Dout_0(FLSIZE-1 downto 0) <= Fout;
	fl_ieee <= Flags(20);

	S1Mode <= (0 => SIMD(0),
		 1 => SIMD(1),
		 2 => SIMD(2),
		 Mode_Sub => Substract);
		 
		S1Fa <= Fa;
		S1Fb <= Fb;
  -- stage 1 : trivial case checks
	stage_1 : process (Clk, Rst, En)
	begin
		
    if (Rst = '1') then
   	  S1out <= '0';
    else
 			if (rising_edge(Clk)) then
      	if (En = '1') then
      	  S1out <= '0';
  			  S1FExout <= (others => '0');
  			  -- TOCKECK : smarter comparaisons?
  			  -- TOADD   : trivial computations?
  			  if (S1Fa = const32_0 or S1Fa = const32_m_0) then               --   0 + ?, 0 + -0
  			    if (S1Fb = const32_0 or S1Fb = const32_m_0) then             --  -0 + 0, 0 +  0
     			    S1Fout <= const32_0;
     			    S1FExout(Ex_ZERO) <= '1';
     			    S1out <= '1';
     			  else
     			  	S1Fout <= S1Fb; -- TODO : check if not NaN or other for S1FExout vector!
     			    S1out <= '1';
     			  end if;
     			elsif (S1Fb = const32_0 or S1Fb = const32_m_0) then            -- ? + 0, ? + -0
     			  S1Fout <= S1Fb;
     			  S1FExout(Ex_ZERO) <= '1';
     			  S1out <= '1';			
   			  elsif (S1Fa = const32_SNaN or S1Fb = const32_SNaN) then        --  SNaN + ?, ? + SNaN
						S1Fout <= const32_SNaN;
						S1FExout(Ex_SNAN) <= '1';
     			  S1out <= '1';
   			  elsif (S1Fa = const32_QNaN or S1Fb = const32_QNaN) then        --  QNaN + ?, ? + QNaN
						S1Fout <= const32_QNaN;
						S1FExout(Ex_QNAN) <= '1';
     			  S1out <= '1';
					elsif (Substract = '0') then 
					  if (S1Fa = const32_m_infty and S1Fb = const32_m_infty) then  --  -infty + -infty
 						  S1Fout <= const32_m_infty;
 						  S1FExout(Ex_INF) <= '1';
     			    S1out <= '1';
	  				elsif (S1Fa = const32_m_infty and S1Fb = const32_infty) then --  -infty + +infty
					    S1Fout <= const32_QNaN;
					    S1FExout(Ex_QNAN) <= '1';
     			    S1out <= '1';
 						elsif (S1Fa = const32_infty and S1Fb = const32_m_infty) then --  +infty + -infty
 						  S1Fout <= const32_QNaN;
 						  S1FExout(Ex_QNAN) <= '1';
     			    S1out <= '1';
 						elsif (S1Fa = const32_infty and S1Fb = const32_infty) then   --  +infty + +infty
 						  S1Fout <= const32_infty;
 						  S1FExout(Ex_INF) <= '1';
     			    S1out <= '1';
 						else
 						  -- not a trivial case : go to stage 2
 							S2Fa <= S1Fa;
 							S2Fb <= S1Fb;
 							S2Mode <= S1Mode;
    				  S1out <= '0';  -- the result is not ready, so stage 2 has to make its job
 						end if;
					else  -- if (Substract = '1') then 
					  if (S1Fa = const32_m_infty and S1Fb = const32_m_infty) then  --  -infty - -infty
 						  S1Fout <= const32_QNaN;
 						  S1FExout(Ex_QNAN) <= '1';
     			    S1out <= '1';
	  				elsif (S1Fa = const32_m_infty and S1Fb = const32_infty) then --  -infty - +infty
					    S1Fout <= const32_m_infty;
					    S1FExout(Ex_INF) <= '1';
     			    S1out <= '1';
 						elsif (S1Fa = const32_infty and S1Fb = const32_m_infty) then --  +infty - -infty
 						  S1Fout <= const32_infty;
 						  S1FExout(Ex_INF) <= '1';
     			    S1out <= '1';
 						elsif (S1Fa = const32_infty and S1Fb = const32_infty) then   --  +infty - +infty
 						  S1Fout <= const32_QNaN;
 						  S1FExout(Ex_QNAN) <= '1';
     			    S1out <= '1';
						else
						  -- not a trivial case : go to stage 2
 							S2Fa <= S1Fa;
 							S2Fb <= S1Fb;
 							S2Mode <= S1Mode;
    				  S1out <= '0';
 						end if;
					end if;
	      end if;
	    end if;
    end if;
  end process;


 	S2Ea <= S2Fa(FLSIZE-1-1 downto FLSIZE-1-1-ESIZE+1);
	S2Eb <= S2Fb(FLSIZE-1-1 downto FLSIZE-1-1-ESIZE+1);

	-- stage 2 : exponent substraction
  stage_2 : process(Clk, Rst, En)
	begin
    if (Rst = '1') then
    	S2ExpAddEn <= '0';
    else
 			if (rising_edge(Clk)) then
 			  -- S2ExpAddEn <= '0';
      	if (En = '1') then
      		if (S1out = '1') then -- stage 1 already has the result
      			S2Fout <= S1Fout;
      			S2FExout <= S1FExout;
      			S2out <='1';
      		else
      			S2out <= '0';
						S2EaF(ESIZE-1 downto 0) <= S2Ea;
						S2EbF(ESIZE-1 downto 0) <= S2Eb;
						
						S2ExpAddEn <= '1'; -- start exponent add/sub
						
						S3Fa <= S2Fa;
						S3Fb <= S2Fb;
						S3Mode <= S2Mode;
					end if;
	      end if;
      end if;
    end if;
	end process;
	

  -- S3DE comes directly from the output of S2ExpAdder (IAdd)
  -- S3DE = Ea - Eb
  -- adding MSB for mantissa
  S3Ma(MSIZE) <= '1';
  S3Mb(MSIZE) <= '1';
  S3Ma(MSIZE-1 downto 0) <= S3Fa(MSIZE-1 downto 0);
  S3Mb(MSIZE-1 downto 0) <= S3Fb(MSIZE-1 downto 0);

	-- stage 3 : mantissa alignment    				
  stage_3 : process(Clk, Rst, En)
    variable S3MaP : std_ulogic_vector (FLSIZE-1 downto 0);
  	variable S3MbP : std_ulogic_vector (FLSIZE-1 downto 0);
    variable S3E   : std_ulogic_vector (ESIZE-1 downto 0);
    variable DEP   : std_ulogic_vector(2 downto 0);
  begin
    if (Rst = '1') then
    else
 			if (rising_edge(Clk)) then    
      	if (En = '1') then
      		if (S2out = '1') then -- stage 2 already has the result
      			S3Fout <= S2Fout;
      			S3FExout <= S2FExout;
      			S3out <='1';
      		else
      			S3out <='0';
      			if (S3DE(ESIZE-1) = '0') then -- DE > 0 => Ea > Eb
	       		  -- Mb right shift
	       		  S3MbP(FLSIZE-1 downto MSIZE+1) := (others => '0');
	       		  S3MbP(MSIZE downto 0) := fasu_rshift(S3Mb(MSIZE downto 0), S3DE(2 downto 0));
	       		  S3MaP(FLSIZE-1 downto MSIZE+1) := (others => '0');
	       		  S3MaP(MSIZE downto 0) := S3Ma(MSIZE downto 0);
	       		  S3E := S3Fa(FLSIZE-1-1 downto FLSIZE-1-1-ESIZE+1);
		        else
		          -- Ea < Eb
		          -- Ma right shift
		          -- DE = -DE
		          DEP := fl_neg(S3DE(2 downto 0));
	       		  S3MaP(FLSIZE-1 downto MSIZE+1) := (others => '0');
	       		  S3MaP(MSIZE downto 0) := fasu_rshift(S3Ma(MSIZE downto 0), DEP);
	       		  S3MbP(FLSIZE-1 downto MSIZE+1) := (others => '0');
	       		  S3MbP(MSIZE downto 0) := S3Mb(MSIZE downto 0);
	       		  S3E := S3Fb(FLSIZE-1-1 downto FLSIZE-1-1-ESIZE+1);
		        end if;
	     		  S4Mode <= S3Mode;
	       		S4Sa <= S3Fa(FLSIZE-1);
	       		S4Sb <= S3Fb(FLSIZE-1);
	       		if (S3Fa(FLSIZE-1) = '1') then -- if Fa < 0 then neg Ma
	       		  S4Ma <= fl_neg(S3MaP);
	       		else        		
	         		S4Ma <= S3MaP;
	       		end if;
	      		if (S3Fb(FLSIZE-1) = '1') then -- if Fb < 0 then neg Mb
	       		  S4Mb <= fl_neg(S3MbP);
	       		else        		
	         		S4Mb <= S3MbP;
	       		end if;
	       		S4E  <= S3E;
		      end if;
		    end if;
	    end if;
    end if;
	end process;
	
	S4AddSub<=S4Mode(Mode_Sub);

	-- stage 4 : mantissa add/sub
	stage_4 : process(Clk, Rst, En)
	begin
    if (Rst = '1') then
    	S4ManAddEn <= '0';
    else
 			if (rising_edge(Clk)) then
 				S4ManAddEn <= '0';
  	    if (En = '1') then
      		if (S3out = '1') then -- stage 1 already has the result
      			S4Fout <= S3Fout;
      			S4FExout <= S3FExout;
      			S4out <='1';
      		else
      			S4out <='0';
					  S4MaF(FLSIZE-1 downto 0) <= S4Ma;
						S4MbF(FLSIZE-1 downto 0) <= S4Mb;
	  			  -- configuring Mantissa Add/Sub
	 			    S4ManAddEn <= '1'; -- lauching adder
	 			    
	 			    S5E <= S4E;
	 			    S5Mode <= S4Mode;
	 			   end if;
			  end if;   
	    end if;    
    end if;	
	end process;
	
	-- note: S5M is given by the result of S4ManAdd (IAdd)
	-- S5M = aligned mantissa 1 +/- aligned mantissa 2
	-- stage5 : normalization and output
	stage_5 : process(Clk, Rst, En)
	  variable S5MP : std_ulogic_vector(MSIZE+1 downto 0);
  	variable S5Mo : std_ulogic_vector(MSIZE-1 downto 0);
  	variable S5EP : std_ulogic_vector(ESIZE-1 downto 0);
  	variable S5Eo : std_ulogic_vector(ESIZE-1 downto 0);
  	variable S5So : std_ulogic;
	begin
    if (Rst = '1') then
			--S7start <= '0';
    else
 			if (rising_edge(Clk)) then
  	    if (En = '1') then
      		if (S4out = '1') then -- stage 1 already has the result
      			S5Fout <= S4Fout;
      			S5FExout <= S4FExout;
      			--S5out <='1';
      		else
      			--S5out <='0';
	  			  --S5start <= '0';
	  			  S5FExout <= (others => '0');	  
				    S5So := S5M(MSIZE + 2);
	  			  if (S5M(MSIZE + 2) = '1') then -- if result mantissa < 0
		  			  S5MP(MSIZE+1 downto 0) := fl_neg(S5M(MSIZE+1 downto 0));
		  			  S5So := '1';  -- negative output
		  			elsif (S5M(MSIZE + 2) = '0') then
		  				S5MP(MSIZE+1 downto 0) := S5M(MSIZE+1 downto 0);
		  				S5So := '0';	  				
	  			  end if;
	  			  if (S5MP(MSIZE + 1) = '1') then -- result mantissa has an overflow (Fa, Fb same sign for addition)
	  			                                  -- ie, the MSB has moved
	                                          -- so: juste right shift by 1
	                                          -- note: loose precision
	            S5Mo(MSIZE-1 downto 0) := S5MP(MSIZE downto 1);
	            -- note: expected: S5MP(MSIZE downto 0) <= S5M(MSIZE+1 downto 1);
	            -- but we remove in the same step the MSB which is now '1'
	            S5Eo := fl_incr(S5E);
	          elsif (S5MP(MSIZE + 1) = '0') then 
	               -- higher bits are NULL so we have to left shift until
	               -- the MSB (bit number MSIZE) become '1'
	               -- occurs when Fa and Fb doesn't have same sign and |Fa| < |Fb|
	            S5EP := S5E;
	            if (fl_isnull(S5MP) = false) then
		            while (S5MP(MSIZE) /= '1') loop
		              S5EP := fl_decr(S5EP);
		             	S5MP(MSIZE downto 1) := S5MP(MSIZE-1 downto 0);
		             	S5MP(0) := '0';
		            end loop;
		          else -- null mantissa
		          end if;
	            S5Eo := S5E;
	            S5Mo := S5MP(MSIZE-1 downto 0); --removing MSB
	  			  end if;
    			  S5Fout(FLSIZE-1) <= S5So;
	  			  S5Fout(FLSIZE-1-1 downto MSIZE) <= S5Eo;
	  			  S5Fout(MSIZE-1 downto 0) <= S5Mo;
	  			end if;
			  end if;   
	    end if;    
    end if;	
	end process;
	
	Fout <= S5Fout;
	FExout <= S5FExout;
	
end Behav_1;



















